"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TemplateLanguageServiceProxy = void 0;
class TemplateLanguageServiceProxy {
    constructor(customizedLanguageServie) {
        this.customizedLanguageServie = customizedLanguageServie;
        this._wrappers = [];
        this.tryAdaptGetApplicableRefactors();
        this.tryAdaptGetEditsForRefactor();
    }
    decorate(languageService) {
        const intercept = Object.create(null);
        for (const { name, wrapper } of this._wrappers) {
            intercept[name] = wrapper(languageService[name].bind(languageService));
        }
        languageService.getApplicableRefactors;
        return new Proxy(languageService, {
            get: (target, property) => {
                return intercept[property] || target[property];
            }
        });
    }
    tryAdaptGetApplicableRefactors() {
        this.wrap('getApplicableRefactors', delegate => (fileName, positionOrRange, preferences, triggerReason) => {
            const original = delegate(fileName, positionOrRange, preferences, triggerReason);
            const customized = this.customizedLanguageServie.getApplicableRefactors(fileName, positionOrRange, preferences, triggerReason);
            return original.concat(customized);
        });
    }
    tryAdaptGetEditsForRefactor() {
        this.wrap('getEditsForRefactor', delegate => (fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) => {
            const result = this.customizedLanguageServie.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
            if (result) {
                return result;
            }
            return delegate(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
        });
    }
    wrap(name, wrapper) {
        this._wrappers.push({ name, wrapper });
        return this;
    }
}
exports.TemplateLanguageServiceProxy = TemplateLanguageServiceProxy;
//# sourceMappingURL=decorator.js.map