import type * as ts from 'typescript/lib/tsserverlibrary';
export declare enum RefactorKind {
    useMemo = 0,
    useCallback = 1
}
declare type MustHaveBody<T extends {
    body?: unknown;
}> = T extends {
    body?: infer B | undefined;
} ? T & {
    body: B;
} : T;
export declare type FunctionExpressionLike = MustHaveBody<ts.ArrowFunction | ts.FunctionDeclaration | ts.FunctionExpression>;
export declare type DependExpression = ts.Identifier | ts.ElementAccessExpression | ts.PropertyAccessExpression;
export interface UseCallbackInfo {
    kind: RefactorKind.useCallback;
    func: FunctionExpressionLike;
    deps: DependExpression[];
    hooksReference: HooksReferenceNameType | undefined;
}
export interface UseMemoInfo {
    kind: RefactorKind.useMemo;
    expression: ts.Expression;
    deps: DependExpression[];
    hooksReference: HooksReferenceNameType | undefined;
}
export declare type Info = UseCallbackInfo | UseMemoInfo;
export interface RefactorContext {
    program: ts.Program;
    file: ts.SourceFile;
}
export declare namespace JsxNames {
    const JSX = "JSX";
    const Element = "Element";
}
export declare const enum Ternary {
    False = 0,
    Unknown = 1,
    Maybe = 3,
    True = -1
}
export declare enum HooksNameType {
    useIdentifier = 0,
    usePropertyAccess = 1
}
export interface IdentifierHooksName {
    type: HooksNameType.useIdentifier;
    alreadyHasReference: boolean;
}
export interface PropertyAccessHooksName {
    type: HooksNameType.usePropertyAccess;
    name: string;
}
export declare type HooksReferenceNameType = IdentifierHooksName | PropertyAccessHooksName;
export {};
