"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.dummyDeDuplicateDeps = exports.skipTriviaExpression = exports.generateImportReactDeclaration = exports.insertNameIntoImportDeclaration = exports.findAndInsertReactImport = exports.createHooksReference = exports.getHooksNameReferenceType = exports.skipSymbolAlias = exports.dummyCheckReactHooks = exports.alreadyWrappedOrContainsInReactHooks = exports.isReactHooks = exports.skipParenthesesUp = exports.wrapIntoJsxExpressionIfNeed = exports.isExpression = exports.skipJsxExpression = exports.skipJsxTextToken = exports.skipSingleValueDeclaration = exports.cloneDeep = exports.isFunctionExpressionLike = exports.createDepSymbolResolver = exports.isDeclarationDefinitelyConstants = exports.isDeclarationAssignedByUseState = exports.isDeclarationAssignedByUseRef = exports.isDeclarationAssignedByReactHooks = exports.isTopLevelConstantDeclaration = exports.isConstantDeclarationIfRoot = exports.byPassRootDeclaration = exports.functionExpressionLikeToExpression = exports.isDefinitelyNotSupportedToken = exports.isInTagNameExpression = exports.isInFunctionComponent = exports.maybeCustomHooks = exports.isFunctionComponentLike = exports.getGlobalJsxElementType = exports.getGlobalJsxElementSymbol = exports.getGlobalJsxNamespace = exports.findTopLevelNodeInSelection = exports.startEndContainsStartEndSkipTrivia = exports.getRangeOfPositionOrRange = void 0;
const types_1 = require("./types");
const helper_1 = require("./helper");
const constants_1 = require("./constants");
function getRangeOfPositionOrRange(positionOrRange) {
    const [startPosition, endPosition] = typeof positionOrRange === 'number'
        ? [positionOrRange, undefined]
        : [positionOrRange.pos, positionOrRange.end];
    return [startPosition, endPosition];
}
exports.getRangeOfPositionOrRange = getRangeOfPositionOrRange;
function startEndContainsStartEndSkipTrivia(typescript, start, end, node, file) {
    return (start <= typescript.skipTrivia(file.text, node.pos) && end >= node.end);
}
exports.startEndContainsStartEndSkipTrivia = startEndContainsStartEndSkipTrivia;
function findTopLevelNodeInSelection(typescript, token, start, end, file) {
    return typescript.findAncestor(token, node => {
        if (!startEndContainsStartEndSkipTrivia(typescript, start, end, node.parent, file)) {
            return true;
        }
        return false;
    });
}
exports.findTopLevelNodeInSelection = findTopLevelNodeInSelection;
function getGlobalJsxNamespace(typescript, checker) {
    return checker.resolveName(types_1.JsxNames.JSX, undefined, typescript.SymbolFlags.Namespace, false);
}
exports.getGlobalJsxNamespace = getGlobalJsxNamespace;
function getSymbol(typescript, checker, symbols, name, meaning) {
    if (meaning) {
        const symbol = checker.getMergedSymbol(symbols.get(name));
        if (symbol) {
            if (symbol.flags & meaning) {
                return symbol;
            }
            if (symbol.flags & typescript.SymbolFlags.Alias) {
                const target = checker.getAliasedSymbol(symbol);
                if (checker.isUnknownSymbol(target) || target.flags & meaning) {
                    return symbol;
                }
            }
        }
    }
    return undefined;
}
function getGlobalJsxElementSymbol(typescript, checker) {
    const globalJsxNamespace = getGlobalJsxNamespace(typescript, checker);
    if (!globalJsxNamespace)
        return undefined;
    const exportsSymbols = checker.getExportsOfModule(globalJsxNamespace);
    const symbolTable = typescript.createSymbolTable(exportsSymbols);
    const elementSymbol = getSymbol(typescript, checker, symbolTable, types_1.JsxNames.Element, typescript.SymbolFlags.Type);
    return elementSymbol;
}
exports.getGlobalJsxElementSymbol = getGlobalJsxElementSymbol;
function getGlobalJsxElementType(typescript, checker) {
    const elementSymbol = getGlobalJsxElementSymbol(typescript, checker);
    return elementSymbol
        ? checker.getDeclaredTypeOfSymbol(elementSymbol)
        : undefined;
}
exports.getGlobalJsxElementType = getGlobalJsxElementType;
function isFunctionComponentLike(node, checker, globalJsxElementType) {
    const signature = checker.getSignatureFromDeclaration(node);
    if (signature) {
        const returnType = checker.getReturnTypeOfSignature(signature);
        if (checker.isTypeAssignableTo(returnType, globalJsxElementType)) {
            return true;
        }
    }
    return false;
}
exports.isFunctionComponentLike = isFunctionComponentLike;
function maybeCustomHooks(typescript, node) {
    switch (node.kind) {
        case typescript.SyntaxKind.FunctionExpression:
        case typescript.SyntaxKind.FunctionDeclaration:
            if (node.name) {
                return helper_1.isUseSomething(node.name.text);
            }
        case typescript.SyntaxKind.ArrowFunction:
            const parent = node.parent;
            if (typescript.isVariableDeclaration(parent) &&
                typescript.isIdentifier(parent.name)) {
                return helper_1.isUseSomething(parent.name.text);
            }
        default:
            return false;
    }
}
exports.maybeCustomHooks = maybeCustomHooks;
function isInFunctionComponent(typescript, node, checker) {
    const globalJsxElementType = getGlobalJsxElementType(typescript, checker);
    if (!globalJsxElementType)
        return 3 /* Maybe */;
    const maybeFC = typescript.findAncestor(node.parent, parent => {
        if (isFunctionExpressionLike(typescript, parent)) {
            return (maybeCustomHooks(typescript, parent) ||
                isFunctionComponentLike(parent, checker, globalJsxElementType));
        }
        if (typescript.isClassLike(parent)) {
            return 'quit';
        }
        return false;
    });
    return maybeFC ? -1 /* True */ : 0 /* False */;
}
exports.isInFunctionComponent = isInFunctionComponent;
function isInTagNameExpression(typescript, node) {
    let lastNode = undefined;
    return !!typescript.findAncestor(node, parent => {
        if (lastNode &&
            (typescript.isJsxOpeningElement(parent) ||
                typescript.isJsxClosingElement(parent)) &&
            typescript.isJsxTagNameExpression(lastNode) &&
            parent.tagName === lastNode) {
            return true;
        }
        lastNode = parent;
        return false;
    });
}
exports.isInTagNameExpression = isInTagNameExpression;
function isDefinitelyNotSupportedToken(typescript, node) {
    switch (node.kind) {
        case typescript.SyntaxKind.JsxOpeningFragment:
        case typescript.SyntaxKind.JsxClosingFragment:
        case typescript.SyntaxKind.JsxOpeningElement:
        case typescript.SyntaxKind.JsxClosingElement:
        case typescript.SyntaxKind.JsxText:
            return true;
        default:
            return isInTagNameExpression(typescript, node);
    }
}
exports.isDefinitelyNotSupportedToken = isDefinitelyNotSupportedToken;
function functionExpressionLikeToExpression(typescript, func) {
    switch (func.kind) {
        case typescript.SyntaxKind.FunctionDeclaration:
            return typescript.factory.createFunctionExpression(func.modifiers, func.asteriskToken, func.name, func.typeParameters, func.parameters, func.type, func.body);
        default:
            return func;
    }
}
exports.functionExpressionLikeToExpression = functionExpressionLikeToExpression;
function byPassRootDeclaration(typescript, node) {
    switch (node.kind) {
        case typescript.SyntaxKind.VariableDeclaration:
        case typescript.SyntaxKind.VariableDeclarationList:
        case typescript.SyntaxKind.BindingElement:
        case typescript.SyntaxKind.ArrayBindingPattern:
        case typescript.SyntaxKind.ObjectBindingPattern:
            return byPassRootDeclaration(typescript, node.parent);
        default:
            return node;
    }
}
exports.byPassRootDeclaration = byPassRootDeclaration;
function isConstantDeclarationIfRoot(typescript, node) {
    switch (node.kind) {
        case typescript.SyntaxKind.VariableStatement:
            return !!(typescript.getCombinedNodeFlags(node.declarationList) & typescript.NodeFlags.Const);
        case typescript.SyntaxKind.FunctionDeclaration:
        case typescript.SyntaxKind.ClassDeclaration:
            return true;
        default:
            return false;
    }
}
exports.isConstantDeclarationIfRoot = isConstantDeclarationIfRoot;
function isTopLevelConstantDeclaration(typescript, declaration) {
    const rootDeclaration = byPassRootDeclaration(typescript, declaration);
    if (typescript.isSourceFile(rootDeclaration.parent)) {
        return isConstantDeclarationIfRoot(typescript, rootDeclaration);
    }
    return false;
}
exports.isTopLevelConstantDeclaration = isTopLevelConstantDeclaration;
function isDeclarationAssignedByReactHooks(typescript, declaration, pred = helper_1.isUseSomething) {
    if (typescript.isVariableDeclaration(declaration) &&
        declaration.initializer &&
        typescript.isCallExpression(declaration.initializer)) {
        if (dummyCheckReactHooks(typescript, declaration.initializer.expression, pred)) {
            return true;
        }
    }
    return false;
}
exports.isDeclarationAssignedByReactHooks = isDeclarationAssignedByReactHooks;
function isDeclarationAssignedByUseRef(typescript, declaration) {
    return isDeclarationAssignedByReactHooks(typescript, declaration, helper_1.isUseRef);
}
exports.isDeclarationAssignedByUseRef = isDeclarationAssignedByUseRef;
function isDeclarationAssignedByUseState(typescript, declaration) {
    var _a;
    if (typescript.isBindingElement(declaration) &&
        typescript.isArrayBindingPattern(declaration.parent) &&
        declaration.parent.elements.length === 2 &&
        declaration.parent.elements[1] === declaration &&
        typescript.isVariableDeclaration(declaration.parent.parent)) {
        const name = (_a = declaration.propertyName) !== null && _a !== void 0 ? _a : declaration.name;
        return (typescript.isIdentifier(name) &&
            helper_1.startsWithIgnoreCase(name.text, constants_1.Constants.SetPrefix) &&
            isDeclarationAssignedByReactHooks(typescript, declaration.parent.parent, helper_1.isUseState));
    }
    return false;
}
exports.isDeclarationAssignedByUseState = isDeclarationAssignedByUseState;
function isDeclarationDefinitelyConstants(typescript, declaration) {
    if (typescript.isEnumDeclaration(declaration) ||
        typescript.isEnumMember(declaration)) {
        return true;
    }
    if (typescript.isVariableDeclaration(declaration) &&
        typescript.getCombinedNodeFlags(declaration) &
            typescript.NodeFlags.Const &&
        declaration.initializer &&
        typescript.isLiteralExpression(declaration.initializer)) {
        return true;
    }
    return;
}
exports.isDeclarationDefinitelyConstants = isDeclarationDefinitelyConstants;
function createDepSymbolResolver(typescript, scope, additionalScope, file, checker) {
    const cached = new Map();
    const textCached = new Set();
    return {
        shouldSymbolDefinitelyBeIgnoreInDeps,
        alreadyDuplicated
    };
    function shouldSymbolDefinitelyBeIgnoreInDeps(rawSymbol) {
        check: if (!cached.has(rawSymbol)) {
            const symbol = skipSymbolAlias(typescript, rawSymbol, checker);
            const valueDeclaration = symbol.valueDeclaration;
            if (!valueDeclaration) {
                cached.set(rawSymbol, false);
                break check;
            }
            if (valueDeclaration.getSourceFile() !== file) {
                cached.set(rawSymbol, true);
                break check;
            }
            if (typescript.rangeContainsRange(scope, valueDeclaration)) {
                cached.set(rawSymbol, true);
                break check;
            }
            if (additionalScope.some(s => typescript.rangeContainsRange(s, valueDeclaration))) {
                cached.set(rawSymbol, true);
                break check;
            }
            if (isDeclarationAssignedByUseRef(typescript, valueDeclaration) ||
                isDeclarationAssignedByUseState(typescript, valueDeclaration)) {
                cached.set(rawSymbol, true);
                break check;
            }
            if (isDeclarationDefinitelyConstants(typescript, valueDeclaration)) {
                cached.set(rawSymbol, true);
                break check;
            }
            if (isTopLevelConstantDeclaration(typescript, valueDeclaration)) {
                cached.set(rawSymbol, true);
                break check;
            }
            cached.set(rawSymbol, false);
        }
        return cached.get(rawSymbol);
    }
    function alreadyDuplicated(symbol) {
        return cached.has(symbol);
    }
}
exports.createDepSymbolResolver = createDepSymbolResolver;
function isFunctionExpressionLike(typescript, v) {
    return ((typescript.isArrowFunction(v) ||
        typescript.isFunctionDeclaration(v) ||
        typescript.isFunctionExpression(v)) &&
        !!v.body);
}
exports.isFunctionExpressionLike = isFunctionExpressionLike;
function cloneDeep(typescript, v) {
    return typescript.transform(v, [
        conext => {
            return node => visitor(node);
            function visitor(node) {
                return typescript.setTextRange(typescript.factory.cloneNode(typescript.visitEachChild(node, visitor, conext)), {
                    pos: -1,
                    end: -1
                });
            }
        }
    ]).transformed[0];
}
exports.cloneDeep = cloneDeep;
function skipSingleValueDeclaration(typescript, node) {
    const original = node;
    if (typescript.isVariableStatement(node)) {
        node = node.declarationList;
    }
    if (typescript.isVariableDeclarationList(node) &&
        node.declarations.length === 1) {
        node = helper_1.first(node.declarations);
    }
    if (typescript.isVariableDeclaration(node) && node.initializer) {
        return node.initializer;
    }
    return original;
}
exports.skipSingleValueDeclaration = skipSingleValueDeclaration;
function skipJsxTextToken(typescript, node, file) {
    if (typescript.isJsxText(node) && node.containsOnlyTriviaWhiteSpaces) {
        const nextToken = typescript.getTokenAtPosition(file, node.end);
        if (nextToken &&
            typescript.rangeContainsRange(node.parent, nextToken)) {
            return nextToken;
        }
    }
    return node;
}
exports.skipJsxTextToken = skipJsxTextToken;
function skipJsxExpression(typescript, node) {
    return (typescript.isJsxExpression(node) && node.expression) || node;
}
exports.skipJsxExpression = skipJsxExpression;
function isExpression(typescript, node) {
    return typescript.isExpressionNode(node);
}
exports.isExpression = isExpression;
function wrapIntoJsxExpressionIfNeed(typescript, node, newNode) {
    return typescript.isJsxElement(node.parent)
        ? typescript.factory.createJsxExpression(undefined, newNode)
        : newNode;
}
exports.wrapIntoJsxExpressionIfNeed = wrapIntoJsxExpressionIfNeed;
function skipParenthesesUp(typescript, node) {
    while (node.kind === typescript.SyntaxKind.ParenthesizedExpression) {
        node = node.parent;
    }
    return node;
}
exports.skipParenthesesUp = skipParenthesesUp;
function alreadyWrappedInReactHooks(typescript, node, checker) {
    const maybeHooks = typescript.findAncestor(node, parent => {
        if (!typescript.isCallExpression(parent) ||
            !typescript.rangeContainsRange(parent.arguments, node)) {
            return false;
        }
        const expression = typescript.skipParentheses(parent.expression);
        const dummyCheckResult = dummyCheckReactHooks(typescript, expression);
        const symbol = checker.getSymbolAtLocation(expression);
        if (!dummyCheckResult || !symbol) {
            return dummyCheckResult;
        }
        return dummyCheckResult && !!isReactHooks(typescript, symbol, checker);
    });
    return !!maybeHooks;
}
function isReactHooks(typescript, symbol, checker) {
    if (!symbol.valueDeclaration) {
        return 3 /* Maybe */;
    }
    const globalJsxElementSymbol = getGlobalJsxElementSymbol(typescript, checker);
    if (!globalJsxElementSymbol) {
        return 3 /* Maybe */;
    }
    const symbolDeclarationFileName = symbol.valueDeclaration.getSourceFile()
        .fileName;
    const symbolPackageName = helper_1.getPackageNameOrNamespaceInNodeModules(symbolDeclarationFileName);
    if (!symbolPackageName) {
        return 0 /* False */;
    }
    const inSamePackage = globalJsxElementSymbol.declarations.some(decl => {
        const fileName = decl.getSourceFile().fileName;
        if (fileName === symbolDeclarationFileName) {
            return true;
        }
        if (helper_1.relativePathContainNodeModules(symbolDeclarationFileName, fileName)) {
            return false;
        }
        const jsxElementPackageName = helper_1.getPackageNameOrNamespaceInNodeModules(fileName);
        if (!jsxElementPackageName) {
            return false;
        }
        if (jsxElementPackageName === symbolPackageName) {
            return true;
        }
    });
    return inSamePackage ? -1 /* True */ : 0 /* False */;
}
exports.isReactHooks = isReactHooks;
function alreadyContainsReactHooks(typescript, node, checker) {
    let maybeHooks = false;
    visitor(node);
    return !!maybeHooks;
    function visitor(child) {
        if (typescript.isCallExpression(child)) {
            const expression = typescript.skipParentheses(child.expression);
            const dummyCheckResult = dummyCheckReactHooks(typescript, expression);
            const symbol = checker.getSymbolAtLocation(expression);
            if (!symbol)
                if (!symbol) {
                    maybeHooks || (maybeHooks = dummyCheckResult);
                    return dummyCheckResult;
                }
                else if (dummyCheckResult &&
                    isReactHooks(typescript, symbol, checker)) {
                    maybeHooks || (maybeHooks = true);
                    return true;
                }
        }
        typescript.forEachChild(child, visitor);
    }
}
function alreadyWrappedOrContainsInReactHooks(typescript, node, checker) {
    return (alreadyWrappedInReactHooks(typescript, node, checker) ||
        alreadyContainsReactHooks(typescript, node, checker));
}
exports.alreadyWrappedOrContainsInReactHooks = alreadyWrappedOrContainsInReactHooks;
function dummyCheckReactHooks(typescript, expression, pred = helper_1.isUseSomething) {
    if (typescript.isIdentifier(expression)) {
        return pred(expression.text);
    }
    if (typescript.isPropertyAccessExpression(expression)) {
        return (typescript.isIdentifier(expression.expression) &&
            helper_1.isReactText(expression.expression.text) &&
            pred(expression.name.text));
    }
    if (typescript.isElementAccessExpression(expression)) {
        return (typescript.isIdentifier(expression.expression) &&
            typescript.isStringLiteralLike(expression.argumentExpression) &&
            helper_1.isReactText(expression.expression.text) &&
            pred(expression.argumentExpression.text));
    }
    return false;
}
exports.dummyCheckReactHooks = dummyCheckReactHooks;
function skipSymbolAlias(typescript, symbol, checker) {
    return symbol.flags & typescript.SymbolFlags.Alias
        ? checker.getAliasedSymbol(symbol)
        : symbol;
}
exports.skipSymbolAlias = skipSymbolAlias;
function getHooksNameReferenceType(typescript, location, checker, hookName, isNewTrasnfromer) {
    const meaning = typescript.SymbolFlags.Value | typescript.SymbolFlags.Alias;
    const identifierSymbol = checker.resolveName(hookName, location, meaning, false);
    const hasIdentifierReactHooksReference = identifierSymbol &&
        isReactHooks(typescript, skipSymbolAlias(typescript, identifierSymbol, checker), checker);
    if (isNewTrasnfromer || hasIdentifierReactHooksReference) {
        return {
            type: types_1.HooksNameType.useIdentifier,
            alreadyHasReference: !!hasIdentifierReactHooksReference
        };
    }
    const upperCaseSymbol = checker.resolveName(constants_1.Constants.React, location, meaning, false);
    if (upperCaseSymbol &&
        isReactHooks(typescript, skipSymbolAlias(typescript, upperCaseSymbol, checker), checker)) {
        return {
            type: types_1.HooksNameType.usePropertyAccess,
            name: constants_1.Constants.React
        };
    }
    return undefined;
}
exports.getHooksNameReferenceType = getHooksNameReferenceType;
function createHooksReference(typescript, sourceFile, hooksReference, hooksName) {
    const factory = typescript.factory;
    if (!hooksReference) {
        return [
            factory.createPropertyAccessExpression(factory.createIdentifier(constants_1.Constants.React), factory.createIdentifier(hooksName)),
            undefined
        ];
    }
    if (hooksReference.type === types_1.HooksNameType.useIdentifier) {
        const postAction = hooksReference.alreadyHasReference
            ? undefined
            : (changeTracker) => {
                findAndInsertReactImport(typescript, sourceFile, hooksName, changeTracker);
            };
        return [factory.createIdentifier(hooksName), postAction];
    }
    return [
        factory.createPropertyAccessExpression(factory.createIdentifier(hooksReference.name), factory.createIdentifier(hooksName)),
        undefined
    ];
}
exports.createHooksReference = createHooksReference;
function findAndInsertReactImport(typescript, sourceFile, hooksName, changeTracker) {
    const importStmt = sourceFile.statements.find(stmt => {
        return (typescript.isImportDeclaration(stmt) &&
            typescript.isStringLiteral(stmt.moduleSpecifier) &&
            stmt.moduleSpecifier.text === constants_1.Constants.ReactModule);
    });
    const reactImportDeclaration = importStmt;
    if (!reactImportDeclaration ||
        !reactImportDeclaration.importClause ||
        !reactImportDeclaration.importClause.namedBindings ||
        !typescript.isNamedImports(reactImportDeclaration.importClause.namedBindings)) {
        changeTracker.insertNodeAtTopOfFile(sourceFile, generateImportReactDeclaration(typescript, hooksName), false);
        return;
    }
    changeTracker.replaceNode(sourceFile, reactImportDeclaration.importClause.namedBindings, insertNameIntoImportDeclaration(typescript, reactImportDeclaration.importClause.namedBindings, hooksName));
}
exports.findAndInsertReactImport = findAndInsertReactImport;
function insertNameIntoImportDeclaration(typescript, namedImports, hooksName) {
    const factory = typescript.factory;
    return factory.updateNamedImports(namedImports, namedImports.elements.concat(factory.createImportSpecifier(undefined, factory.createIdentifier(hooksName))));
}
exports.insertNameIntoImportDeclaration = insertNameIntoImportDeclaration;
function generateImportReactDeclaration(typescript, hooksName) {
    const factory = typescript.factory;
    return factory.createImportDeclaration(undefined, undefined, factory.createImportClause(false, undefined, factory.createNamedImports([
        factory.createImportSpecifier(undefined, factory.createIdentifier(hooksName))
    ])), factory.createStringLiteral(constants_1.Constants.ReactModule));
}
exports.generateImportReactDeclaration = generateImportReactDeclaration;
function skipTriviaExpression(typescript, expression) {
    switch (expression.kind) {
        case typescript.SyntaxKind.AsExpression:
        case typescript.SyntaxKind.ParenthesizedExpression:
            return expression
                .expression;
        default:
            return expression;
    }
}
exports.skipTriviaExpression = skipTriviaExpression;
function dummyDeDuplicateDeps(deps) {
    const dummyTextRecord = new Map();
    deps.forEach(dep => {
        const text = dep.getText().trim();
        if (!dummyTextRecord.has(text)) {
            dummyTextRecord.set(text, dep);
        }
    });
    return Array.from(dummyTextRecord.values());
}
exports.dummyDeDuplicateDeps = dummyDeDuplicateDeps;
//# sourceMappingURL=utils.js.map